#import "FileSystem.h"

#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/event.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <unistd.h>

#import <Foundation/NSArray.h>
#import <Foundation/NSAutoreleasePool.h>
#import <Foundation/NSString.h>
#import <AppKit/NSWorkspace.h>

JNIEXPORT void JNICALL
Java_com_deltopia_internal_io_macosx_FileSystem_close (JNIEnv *jniEnv,
		jobject obj, jint fileDesc) {
	close (fileDesc);
}

JNIEXPORT jint JNICALL
Java_com_deltopia_internal_io_macosx_FileSystem_getUnixMode (JNIEnv *jniEnv,
		jobject obj, jstring inPath, jboolean inFollowLinks, jintArray outUnixMode) {
	const char *path;
	jint status = 0;

	path = (*jniEnv)->GetStringUTFChars (jniEnv, inPath, NULL);
	if (path) {
		jint *unixModePtr;

		if (outUnixMode) {
			unixModePtr =
				(*jniEnv)->GetIntArrayElements (jniEnv, outUnixMode, NULL);
			if (NULL == unixModePtr) {
				status = ENOMEM;
			}
		} else {
			unixModePtr = NULL;
		}
		if (0 == status) {
			struct stat fileInfo;

			status = (JNI_TRUE == inFollowLinks) ? stat (path, &fileInfo) : lstat (path, &fileInfo);
			
			if (-1 == status) {
				status = errno;
			} else if (unixModePtr) {
				*unixModePtr = fileInfo.st_mode; 
			}
			
			if (unixModePtr) {
				(*jniEnv)->ReleaseIntArrayElements (jniEnv, outUnixMode,
					unixModePtr, 0);
			}
		}
		(*jniEnv)->ReleaseStringUTFChars (jniEnv, inPath, path);
	} else {
		status = ENOMEM;
	}
	return status;
}

JNIEXPORT jint JNICALL
Java_com_deltopia_internal_io_macosx_FileSystem_kevent (JNIEnv *jniEnv,
		jobject obj, jint inKQueue, jintArray inKEvent, jintArray outKEvent,
		jint inTimeout) {
	jint status;
	int inKevCount;
	struct kevent inKev;

	if (inKEvent) {
		jint *inKEventPtr;

		inKEventPtr = (*jniEnv)->GetIntArrayElements (jniEnv, inKEvent, NULL);
		if (inKEventPtr) {
			status = 0;
			inKevCount = 1;
			EV_SET(&inKev, inKEventPtr[0], inKEventPtr[1], inKEventPtr[2],
				inKEventPtr[3], inKEventPtr[4], (void *) inKEventPtr[5]);
			(*jniEnv)->ReleaseIntArrayElements (jniEnv, inKEvent, inKEventPtr, 0);
		} else {
			status = ENOMEM;
		}
	} else {
		status = 0;
		inKevCount = 0;
	}
	if (0 == status) {
		jint *outKEventPtr;
		int outKevCount;
		struct kevent outKev;

		if (outKEvent) {
			outKEventPtr =
				(*jniEnv)->GetIntArrayElements (jniEnv, outKEvent, NULL);
			if (outKEventPtr) {
				outKevCount = 1;
			} else {
				status = ENOMEM;
			}
		} else {
			outKEventPtr = NULL;
			outKevCount = 0;
		}
		if (0 == status) {
			struct timespec timeout = { inTimeout, 0 };
			int kevCount =
				kevent (inKQueue, &inKev, inKevCount, &outKev, outKevCount,
					&timeout);

			switch (kevCount) {
			case -1:
				status = errno;
				break;
			case 0:
				if (0 != outKevCount) {

					/* The time limit has expired. */
					outKEventPtr[0] = -1;
					outKEventPtr[1] = 0;
					outKEventPtr[2] = 0;
					outKEventPtr[3] = 0;
					outKEventPtr[4] = 0;
					outKEventPtr[5] = 0;
				}
				break;
			default:
				if (0 == outKevCount) {

					/*
					 * As no kevent is to be retrieved, the execution shouldn't
					 * come here but, if it does, report an error because
					 * there's no knowledge how to handle the situation.
					 */
					status = EINVAL;
				} else {
					outKEventPtr[0] = outKev.ident;
					outKEventPtr[1] = outKev.filter;
					outKEventPtr[2] = outKev.flags;
					outKEventPtr[3] = outKev.fflags;
					outKEventPtr[4] = outKev.data;
					outKEventPtr[5] = (jint) outKev.udata;
				}
				break;
			}
			if (outKEvent && outKEventPtr) {
				(*jniEnv)->ReleaseIntArrayElements (jniEnv, outKEvent,
					outKEventPtr, 0);
			}
		}
	}
	return status;
}

JNIEXPORT jint JNICALL
Java_com_deltopia_internal_io_macosx_FileSystem_kqueue (JNIEnv *jniEnv,
		jobject obj, jintArray outKQueue) {
	jint *outKQueuePtr;
	jint status;

	if (outKQueue) {
		outKQueuePtr = (*jniEnv)->GetIntArrayElements (jniEnv, outKQueue, NULL);
		if (NULL == outKQueuePtr) {

			/* An exception has already been thrown. */
			status = ENOMEM;
		} else {
			status = 0;
		}
	} else {
		outKQueuePtr = NULL;
		status = 0;
	}
	if (0 == status) {
		int kq = kqueue ();
		
		if (-1 == kq) {
			status = errno;
		} else if (outKQueuePtr) {
			*outKQueuePtr = kq;
		}
		if (outKQueue && outKQueuePtr) {
			(*jniEnv)->ReleaseIntArrayElements (jniEnv, outKQueue, outKQueuePtr, 0);
		}
	}
	return status;
}

JNIEXPORT jboolean JNICALL
Java_com_deltopia_internal_io_macosx_FileSystem_moveToTrash (JNIEnv *jniEnv,
		jobject obj, jstring absoluteParentPath, jstring fileName) {
	NSAutoreleasePool *autoreleasePool;
	const jchar *chars;
	NSString *absoluteParentPathAsNSString;
	NSString *fileNameAsNSString;
	BOOL result;
	int tag;
	
	autoreleasePool = [[NSAutoreleasePool alloc] init];

	chars = (*jniEnv)->GetStringChars (jniEnv, absoluteParentPath, NULL);
	absoluteParentPathAsNSString = [NSString
			stringWithCharacters:(const unichar *)chars
			length:(*jniEnv)->GetStringLength (jniEnv, absoluteParentPath)];
	(*jniEnv)->ReleaseStringChars (jniEnv, absoluteParentPath, chars);

	chars = (*jniEnv)->GetStringChars (jniEnv, fileName, NULL);
	fileNameAsNSString = [NSString stringWithCharacters:(const unichar *)chars
			length:(*jniEnv)->GetStringLength (jniEnv, fileName)];
	(*jniEnv)->ReleaseStringChars (jniEnv, fileName, chars);

	result = [[NSWorkspace sharedWorkspace]
			performFileOperation:NSWorkspaceRecycleOperation
			source:absoluteParentPathAsNSString destination:@""
			files:[NSArray arrayWithObjects:fileNameAsNSString, nil]
			tag:&tag] && (tag >= 0);

	[autoreleasePool release];
	return (YES == result) ? JNI_TRUE : JNI_FALSE;
}

JNIEXPORT jint JNICALL
Java_com_deltopia_internal_io_macosx_FileSystem_open (JNIEnv *jniEnv,
		jobject obj, jstring inPath, jint inFlags, jintArray outFileDesc) {
	const char *path;
	jint status = 0;

	path = (*jniEnv)->GetStringUTFChars (jniEnv, inPath, NULL);
	if (path) {
		jint *fileDescPtr;

		if (outFileDesc) {
			fileDescPtr =
				(*jniEnv)->GetIntArrayElements (jniEnv, outFileDesc, NULL);
			if (NULL == fileDescPtr) {

				/* An exception has already been thrown. */
				status = ENOMEM;
			}
		} else {
			fileDescPtr = NULL;
		}
		if (0 == status) {
			int fileDesc;

			fileDesc = open (path, (int) inFlags);
			if (-1 == fileDesc) {
				status = errno;
			}
			if (fileDescPtr) {
				*fileDescPtr = fileDesc;
				(*jniEnv)->ReleaseIntArrayElements (jniEnv, outFileDesc,
					fileDescPtr, 0);
			}
		}
		(*jniEnv)->ReleaseStringUTFChars (jniEnv, inPath, path);
	} else {

		/*
		 * It doesn't matter much what status is returned because an
		 * OutOfMemoryError has already been thrown.
		 */
		status = ENOMEM;
	}
	return status;
}

JNIEXPORT jint JNICALL
Java_com_deltopia_internal_io_macosx_FileSystem_readLink (JNIEnv *jniEnv,
		jobject obj, jstring inPath, jbyteArray outBuffer, jint inBufferSize) {
	const char *path;
	jint retValue = 0;

	path = (*jniEnv)->GetStringUTFChars (jniEnv, inPath, NULL);
	if (path == NULL) {
		return ENOMEM;
	}

	jbyte *bufferPtr;
	if (outBuffer) {
		bufferPtr =
			(*jniEnv)->GetByteArrayElements (jniEnv, outBuffer, NULL);
		if (NULL == bufferPtr) {
			retValue = ENOMEM;
		}
	} else {
		bufferPtr = NULL;
	}

	if (0 == retValue) {
		retValue = readlink (path, (char*)bufferPtr, inBufferSize);
		
		if ((-1 != retValue) && (bufferPtr != NULL)) {
			(*jniEnv)->ReleaseByteArrayElements (jniEnv, outBuffer,
				bufferPtr, 0);
		}
	}
		
	if (-1 == retValue) {
		retValue = errno;
	}
	
	(*jniEnv)->ReleaseStringUTFChars (jniEnv, inPath, path);

	return retValue;
}

JNIEXPORT jint JNICALL
Java_com_deltopia_internal_io_macosx_FileSystem_setUnixMode (JNIEnv *jniEnv,
		jobject obj, jstring path, jint unixMode) {
	const char *pathAsCString;
	jint status;

	pathAsCString = (*jniEnv)->GetStringUTFChars (jniEnv, path, NULL);
	if (pathAsCString) {
		status = chmod (pathAsCString, (mode_t) unixMode);
		if (-1 == status) {
			status = errno;
		}
		(*jniEnv)->ReleaseStringUTFChars (jniEnv, path, pathAsCString);
	} else {
		status = ENOMEM;
	}
	return status;
}
